
from threading import Thread
from pathlib import Path

from pandas import DataFrame
from math import exp
from numpy import power
from corems.molecular_id.factory.EI_SQL import EI_LowRes_SQLite
from corems.molecular_id.calc.SpectralSimilarity import SpectralSimilarity

class LowResMassSpectralMatch(Thread):

    def __init__(self, gcms_obj, sql_obj=None, calibration=False):

        '''TODO:
        '''
        Thread.__init__(self)

        self.gcms_obj = gcms_obj

        #  initiated at create_molecular_database()
        #self.dict_molecular_lookup_table = None
        self.calibration = calibration
        # reading local file for now,
        if not sql_obj:
            self.sql_obj = EI_LowRes_SQLite(url=self.gcms_obj.molecular_search_settings.url_database)
        else:
            self.sql_obj = sql_obj

    def metabolite_detector_score(self, gc_peak, ref_obj, spectral_simi):

        spectral_similarity_scores = {}
        spectral_similarity_scores["cosine_correlation"] = spectral_simi.cosine_correlation()

        if self.gcms_obj.molecular_search_settings.exploratory_mode:

            spectral_similarity_scores["weighted_cosine_correlation"] = spectral_simi.weighted_cosine_correlation()
            ss, ss_nist = spectral_simi.stein_scott()
            spectral_similarity_scores["stein_scott_similarity"] = ss
            spectral_similarity_scores["stein_scott_similarity_nist"] = ss_nist

            spectral_similarity_scores["pearson_correlation"] = spectral_simi.pearson_correlation()
            spectral_similarity_scores["spearman_correlation"] = spectral_simi.spearman_correlation()
            spectral_similarity_scores["kendall_tau_correlation"] = spectral_simi.kendall_tau()
            spectral_similarity_scores["euclidean_distance"] = spectral_simi.euclidean_distance()
            spectral_similarity_scores["manhattan_distance"] = spectral_simi.manhattan_distance()
            spectral_similarity_scores["jaccard_distance"] = spectral_simi.jaccard_distance()
            spectral_similarity_scores["dft_correlation"] = spectral_simi.dft_correlation()
            spectral_similarity_scores["dwt_correlation"] = spectral_simi.dwt_correlation()
            spectral_similarity_scores.update(spectral_simi.extra_distances())
            # print(spectral_similarity_scores)
        # print(ref_obj.get('ri'), gc_peak.ri, self.gcms_obj.molecular_search_settings.ri_window)

        ri_score = exp(-1 * (power((gc_peak.ri - ref_obj.get('ri')), 2) / (2 * power(self.gcms_obj.molecular_search_settings.ri_std, 2))))

        similarity_score = ((spectral_similarity_scores.get("cosine_correlation")**2) * (ri_score))**(1 / 3)

        return spectral_similarity_scores, ri_score, similarity_score

    # @timeit
    def run(self):
        # TODO select the best gcms peak    
        import tqdm

        original_use_deconvolution = self.gcms_obj.chromatogram_settings.use_deconvolution

        if not self.gcms_obj:

            # Do not use deconvolution for the retention index calibration

            if self.calibration:

                self.gcms_obj.chromatogram_settings.use_deconvolution = False

            self.gcms_obj.process_chromatogram()

        self.gcms_obj.chromatogram_settings.use_deconvolution = original_use_deconvolution

        for gc_peak in tqdm.tqdm(self.gcms_obj):
            
            if not self.calibration:
                
                window = self.gcms_obj.molecular_search_settings.ri_search_range

                ri = gc_peak.ri
                
                min_mat_ri = (ri-window, ri+window)    
                
                ref_objs = self.sql_obj.query_min_max_ri(min_mat_ri)
                
            else:

                compound_names = self.gcms_obj.molecular_search_settings.ri_calibration_compound_names

                window = self.gcms_obj.molecular_search_settings.rt_search_range

                rt = gc_peak.rt

                min_mat_rt = (rt-window, rt+window)    
                
                ref_objs = self.sql_obj.query_names_and_rt(min_mat_rt, compound_names)
                
            for ref_obj in ref_objs:
                # uses spectral similarly and uses a threshold to only select peaks with high data correlation
                
                spectral_simi = SpectralSimilarity(gc_peak.mass_spectrum.mz_abun_dict, ref_obj)

                if self.calibration:
                    
                    spectral_similarity_scores = {}
                    spectral_similarity_scores["cosine_correlation"] = spectral_simi.cosine_correlation()
                    
                    #print(w_correlation_value,correlation_value )
                    if spectral_similarity_scores["cosine_correlation"] >= self.gcms_obj.molecular_search_settings.correlation_threshold:
                    
                        gc_peak.add_compound(ref_obj, spectral_similarity_scores)

                # use score, usually a combination of Retention index and Spectral Similarity
                # Threshold is implemented by not necessarily used
                else:

                    # m/q developed methods will be implemented here
                    spectral_similarity_scores, ri_score, similarity_score = self.metabolite_detector_score(gc_peak, ref_obj, spectral_simi)   

                    #TODO need to add similarity score option in the parameters encapsulation class

                    if similarity_score >= self.gcms_obj.molecular_search_settings.score_threshold:

                        gc_peak.add_compound(ref_obj, spectral_similarity_scores, ri_score, similarity_score)


        self.sql_obj.session.close()
        self.sql_obj.engine.dispose()



__author__ = "Yuri E. Corilo"
__date__ = "Oct 24, 2019"

from threading import Thread
from pathlib import Path
import sys, re, json

sys.path.append('.')

from corems.molecular_formula.factory.MolecularFormulaFactory import MolecularFormula 
from corems.encapsulation.constant import Labels
from corems.encapsulation.constant import Atoms
from corems.molecular_id.factory.molecularSQL import CarbonHydrogen, MolecularFormulaLink, HeteroAtoms

class MolecularFormulaLinkProxy():
        
        def __init__(self, molecular_formula, mz):

            self.C = molecular_formula.get('C')
            self.H = molecular_formula.get('H')
            self.H_C = molecular_formula.get('H')/molecular_formula.get('C')
            self.classe = json.dumps(molecular_formula.class_dict)
            self.mass =  float(mz)                       
            self.dbe = molecular_formula.dbe
            self.formula_dict = molecular_formula.to_dict()

class ImportMassListRef():#Thread

    def __init__(self, ref_file_location) :
            
            #Thread.__init__(self)
            
            self.ref_file_location = Path(ref_file_location)
            
            if not self.ref_file_location.exists():
                tb = sys.exc_info()[2]
                raise FileNotFoundError(ref_file_location).with_traceback(tb)
    
    def molecular_formula_ref( self, mz, molecular_formula):
        
        return MolecularFormulaLinkProxy(molecular_formula, mz)
    
    def from_bruker_ref_file(self):

        import csv
        
        list_mf_obj = []

        with open(self.ref_file_location) as ref_f:

            labels = ref_f.readline().strip('\n').split(';')
            
            for line in ref_f.readlines():
                
                if line != '\n':
        
                    list_ref = (line.strip('\n').split(' '))
                    
                    if list_ref[2][-1] == '+': 
                        
                        ion_charge =  int(list_ref[2][:-1])
                    
                    else:
                        
                        ion_charge =  -1* int(list_ref[2][:-1])

                    
                    ion_mol_formula = list_ref[0]
                    mz = list_ref[1]
                    formula_dict = self.mformula_s_to_dict(ion_mol_formula)
                    
                    list_mf_obj.append(self.molecular_formula_ref(mz, MolecularFormula(formula_dict, ion_charge)))
        
        return  list_mf_obj           

    def from_corems_ref_file(self, delimiter="\t"): #pragma: no cover
        '''not being used'''
        import csv

        list_mf_obj = []

        with open('res/RefMassLists/Crude-Pos-ESI.ref') as ref_f:

            labels = ref_f.readline().strip('\n').split(delimiter)
            
            for line in ref_f.readlines():
                
                if line != '\n':
        
                    list_ref = (line.strip('\n').split(delimiter))
                    
                    formula_string = list_ref[0]
                    ion_charge = int(list_ref[1])
                    ion_type = list_ref[2]

                    molform = MolecularFormula(formula_string, ion_charge, ion_type=ion_type)
                    
                    list_mf_obj.append(self.molecular_formula_ref(molform))

        return  list_mf_obj           


    def split(self, delimiters, string, maxsplit=0): #pragma: no cover
    
        ''' does not work when formula has atoms with same caracaters:
            i.e - C10H21NNa
        '''
        regexPattern = '|'.join(map(re.escape, delimiters)) #pragma: no cover
        isotopes = re.findall(regexPattern, string) #pragma: no cover
        counts = re.split(regexPattern, string, maxsplit)  #pragma: no cover
        return isotopes, counts

    def mformula_s_to_dict(self, s_mformulatring):
        
        ''' 
            Converts a molecular formula string to a dict
            
            args:
            
            s_mformulatring: str
                'C10H21NNa'
            
            -   Does not work if the atomic mass number is passed i.e. 37Cl, 81Br
                The convention follow the light isotope labeling 35Cl is Cl, 12C is C, etc
            
           -    If you need to use heavy isotopes please use another reference file format that 
                separate the formula string by a blank space and parse it using the function corems_ref_file""   
        
        '''

        if s_mformulatring:
            
            #find the case C122
            all_atoms = re.findall(r'[A-Z]{1}[0-9]{1,10000}', s_mformulatring)
            
            #find the case Br2
            all_atoms2 = re.findall(r'[A-Z]{1}[a-z]{1}[0-9]{1,10000}', s_mformulatring)

            #find the case N
            single_digit_atoms_one = re.findall(r'[A-Z]{1}(?![0-9])(?![A-Z])', s_mformulatring)

            #find the case Na
            due_digit_atoms_one = re.findall(r'[A-Z]{1}[a-z]{1}(?![0-9])', s_mformulatring)

            all_atoms = all_atoms + all_atoms2 + single_digit_atoms_one + due_digit_atoms_one
            
            dict_res = {}
            
            for each_atom_count in all_atoms:
                
                
                count = re.findall(r'[0-9]{1,10000}', each_atom_count)
                atom = ''.join(re.findall(r'[A-z]', each_atom_count))
                
                if atom in Atoms.atoms_order:
                    
                    if count:
                        dict_res[atom] = int(count[0])
                    else:
                        dict_res[atom] = 1
                
                else:
                    
                    tb = sys.exc_info()[2]
                    raise TypeError("Atom %s does not exist in Atoms.atoms_order list" % atom).with_traceback(tb)
            
            dict_res[Labels.ion_type]  =  'unknown'

            return dict_res
        
        else: 
            
            tb = sys.exc_info()[2]
            raise Exception('Empty molecular formula').with_traceback(tb)

    
    
        
                


from numpy.fft import rfft
from pywt import dwt 
from scipy.spatial.distance import cosine, jaccard, euclidean, cityblock
from scipy.stats import pearsonr, spearmanr, kendalltau
from sklearn.metrics.pairwise import cosine_similarity
from numpy import power, dot, absolute, subtract, intersect1d, where, average, corrcoef, sqrt
from numpy import sum as np_sum
from numpy.linalg import norm
from pandas import DataFrame
import numpy as np

methods_name = {
    #"entropy_distance": "Entropy Distance",
    #"weighted_entropy_distance": "Dynamic weighted entropy Distance",
    "chebyshev_distance": "Chebyshev Distance",
    "squared_euclidean_distance": "Squared Euclidean Distance",
    "fidelity_similarity": "Fidelity Similarity",
    "matusita_distance": "Matusita Distance",
    "squared_chord_distance": "Squared-chord Distance",
    #"bhattacharya_1_distance": "Bhattacharya 1 Distance",
    #"bhattacharya_2_distance": "Bhattacharya 2 Distance",
    "harmonic_mean_similarity": "Harmonic mean Distance",
    "Pearson_chi_squared_distance": "Pearson Chi Squared Distance",
    "Neyman_chi_squared_distance": "Neyman Chi Squared Distance",
    "probabilistic_symmetric_chi_squared_distance": "Probabilistic symmetric X2 Distance",
    "topsoe_distance": "Topsoe Distance",
    "chernoff_distance": "Chernoff Distance",
    "ruzicka_distance": "Ruzicka Distance",
    "roberts_distance": "Roberts Distance",
    #"intersection_distance": "Intersection Distance",
    "motyka_distance": "Motyka Distance",
    "canberra_distance": "Canberra Distance",
    "canberra_metric": "Canberra Metric",
    "kulczynski_1_distance": "Kulczynski 1 Distance",
    #"baroni_urbani_buser_distance": "Baroni-Urbani-Buser Distance",
    #"penrose_size_distance": "Penrose size Distance",
    #"mean_character_distance": "Mean character Distance",
    "lorentzian_distance": "Lorentzian Distance",
    #"penrose_shape_distance": "Penrose shape Distance",
    "clark_distance": "Clark Distance",
    "hellinger_distance": "Hellinger Distance",
    "whittaker_index_of_association_distance": "Whittaker index of association Distance",
    #"similarity_index_distance": "Similarity Index Distance",
    #"improved_similarity_distance": "Improved Similarity",
    #"absolute_value_distance": "Absolute Value Distance",
    "spectral_contrast_angle_distance": "Spectral Contrast Angle",
    "wave_hedges_distance": "Wave Hedges Distance",
    "dice_similarity": "Dice Similarity",
    "inner_product_distance": "Inner Product Distance",
    "divergence_distance": "Divergence Distance",
    "jensen_difference_distance": "Jensen Differences Distance",
    "kumar_johnson_distance": "Kumar Johnson Distance",
    "avg_l_distance": "Avg (L1, L8) Distance",
    "vicis_wave_hadges_distance": "Vicis Wave Hadges Distance",
    "vicis_symmetric_chi_squared_1_distance": "Vicis-Symmetric X2 1 Distance",
    "vicis_symmetric_chi_squared_2_distance": "Vicis-Symmetric X2 2 Distance",
    "vicis_symmetric_chi_squared_3_distance": "Vicis-Symmetric X2 3 Distance",
    "max_symmetric_chi_squared_distance": "Max Symmetric Chi Squared Distance",
    "min_symmetric_chi_squared_distance": "Min Symmetric Chi Squared Distance",
    #"ms_for_id_v1": "MSforID Distance version 1",
    #"ms_for_id": "MSforID Distance",
    "additive_sym_chi_sq": "Additive Symmetric Chi Squared",
    "bhattacharya_distance": "Battacharya Distance",
    "generalized_ochai_index": "Generalized Ochai Index",
    "gower_distance": "Gower Distance",
    "impr_sqrt_cosine_sim": "Improved Square Root Cosine Similarity",
    "intersection_sim": "Intersection Similarity",
    "j_divergence": "J Divergence",
    "jensen_shannon_index": "Jensen Shannon Index",
    "k_divergence": "K Divergence",
    "VW6": "VW6",
    "VW5": "VW5",
    "VW4": "VW4",
    "VW3": "VW3",
    "VW2": "VW2",
    "VW1": "VW1",
    "taneja_divergence": "Taneja Divergence",
    "symmetric_chi_squared_distance": "Symmetric Chi Squared Distance",
    "squared_chi_squared_distance": "Squared Chi Squared Distance",
    "square_root_cosine_correlation": "Square Root Cosine Correlation",
    "sorensen_distance": "Sorensen Distance",
    "Minokowski_3": "Minokowski 3 Distance",
    "Minokowski_4": "Minokowski 4 Distance",
    "kumarjohnson_divergence": "Kumar Johnson Divergence",
    "kumarhassebrook_similarity": "Kumar Hassebrook Similarity",
    "kullbackleibler_divergence": "Kullback Leibler Divergence",
    "soergel_distance": "Soergel Distance",
}

methods_scale = {
    "entropy": [0, np.log(4)],
    "weighted_entropy": [0, np.log(4)],
    "absolute_value": [0, 2],
    "avg_l": [0, 1.5],
    "bhattacharya_1": [0, np.arccos(0) ** 2],
    "bhattacharya_2": [0, np.inf],
    "canberra": [0, np.inf],
    "clark": [0, np.inf],
    "divergence": [0, np.inf],
    "euclidean": [0, np.sqrt(2)],
    "hellinger": [0, np.inf],
    "improved_similarity": [0, np.inf],
    "lorentzian": [0, np.inf],
    "manhattan": [0, 2],
    "matusita": [0, np.sqrt(2)],
    "mean_character": [0, 2],
    "motyka": [-0.5, 0],
    "ms_for_id": [-np.inf, 0],
    "ms_for_id_v1": [0, np.inf],
    "pearson_correlation": [-1, 1],
    "penrose_shape": [0, np.sqrt(2)],
    "penrose_size": [0, np.inf],
    "probabilistic_symmetric_chi_squared": [0, 1],
    "similarity_index": [0, np.inf],
    "squared_chord": [0, 2],
    "squared_euclidean": [0, 2],
    "symmetric_chi_squared": [0, 0.5 * np.sqrt(2)],
    "topsoe": [0, np.sqrt(2)],
    "vicis_symmetric_chi_squared_3": [0, 2],
    "wave_hedges": [0, np.inf],
    "whittaker_index_of_association": [0, np.inf]
}

class SpectralSimilarity():

    def __init__(self, ms_mz_abun_dict, ref_obj, norm_func=sum):

        self.normalize_func = norm_func
        self.ms_mz_abun_dict = ms_mz_abun_dict
        self.ref_obj = ref_obj

        self.exp_abun = list(self.ms_mz_abun_dict.values())
        self.exp_mz = list(self.ms_mz_abun_dict.keys())

        self.ref_mz = self.ref_obj.get("mz")
        self.ref_abun = self.ref_obj.get("abundance")

        self.ref_mz_abun_dict = dict(zip(self.ref_mz, self.ref_abun))

        # parse to dataframe, easier to zerofill and tranpose
        self.df = DataFrame([self.ms_mz_abun_dict, self.ref_mz_abun_dict])

        # fill missing mz with abundance 0
        x, y = self.nan_fill(self.df, fill_with=1e-10)
        
        self.zero_filled_u_l = self.normalize(x, y, norm_func=self.normalize_func)
        
        # filter out the mass values that have zero intensities in self.exp_abun
        exp_mz_filtered = set([k for k in self.exp_mz if self.ms_mz_abun_dict[k] != 0])

        # filter out the mass values that have zero intensities in self.ref_mz
        ref_mz_filtered = set([k for k in self.ref_mz if self.ref_mz_abun_dict[k] != 0])

        # find the intersection/common mass values of both ref and exp, and sort them
        self.common_mz_values = sorted(list(exp_mz_filtered.intersection(ref_mz_filtered)))

        # find the number of common mass values (after filtering 0s)
        self.n_x_y = len(self.common_mz_values)
        # print(self.n_x_y)

    def nan_fill(self, df, fill_with=0):

        df.fillna(fill_with, inplace=True)
        
        return df.T[0].values, df.T[1].values

    def normalize(self, x, y, norm_func=sum):

        u_l = (x / norm_func(x), y / norm_func(y) )

        return u_l

    def weighted_cosine_correlation(self, a=0.5, b=1.3, nanfill=1e-10):

        # create dict['mz'] = abundance, for experimental data
        # ms_mz_abun_dict = mass_spec.mz_abun_dict
        # weight exp data

        xc = power(self.exp_abun, a) * power(self.exp_mz, b)

        # track back to individual mz
        weighted_exp_dict = dict(zip(self.ms_mz_abun_dict.keys(), xc))

        # weight ref data
        yc = power(self.ref_obj.get("abundance"), a) * power(self.ref_obj.get("mz"), b)

        ref_mz_abun_dict = dict(zip(self.ref_obj.get("mz"), yc))

        # parse to dataframe, easier to zerofill and tranpose
        df = DataFrame([weighted_exp_dict, ref_mz_abun_dict])

        # fill missing mz with weight {abun**a}{m/z**b} to 0
        x, y = self.nan_fill(df, fill_with=nanfill)
        
        # correlation = (1 - cosine(x, y))

        correlation = dot(x, y) / (norm(x) * norm(y))

        return correlation

    def cosine_correlation(self):

        # calculate cosine correlation,
        x = self.zero_filled_u_l[0]
        y = self.zero_filled_u_l[1]

        # correlation = (1 - cosine(x, y))

        correlation = dot(x, y) / (norm(x) * norm(y))

        return correlation

    def stein_scott(self):

        if self.n_x_y == 0: return 0, 0

        # count number of non-zero abundance/peak intensity values
        n_x = sum(a != 0 for a in self.exp_abun)

        s_r_x_y = 0

        a, b = 1, 0

        for i in range(1, self.n_x_y):

            current_value = self.common_mz_values[i]
            previous_value = self.common_mz_values[i - 1]

            y_i = self.ref_mz_abun_dict[current_value]
            y_i_minus1 = self.ref_mz_abun_dict[previous_value]

            lc_current = power(y_i, a) * power(current_value, b)
            lc_previous = power(y_i_minus1, a) * power(previous_value, b)

            x_i = self.ms_mz_abun_dict[current_value]
            x_i_minus1 = self.ms_mz_abun_dict[previous_value]

            uc_current = power(x_i, a) * power(current_value, b)
            uc_previous = power(x_i_minus1, a) * power(previous_value, b)

            T1 = lc_current / lc_previous

            T2 = uc_previous / uc_current

            temp_computation = T1 * T2

            n = 0
            if temp_computation <= 1:
                n = 1
            else:
                n = -1
            s_r_x_y = s_r_x_y + power(temp_computation, n)

        # finish the calculation of S_R(X,Y)

        s_r_x_y = s_r_x_y / self.n_x_y
        # using the existing weighted_cosine_correlation function to get S_WC(X,Y)
        s_wc_x_y = self.weighted_cosine_correlation(a=0.5, b=3, nanfill=0)

        s_ss_x_y = ((n_x * s_wc_x_y) + (self.n_x_y * s_r_x_y)) / (n_x + self.n_x_y)

        s_wc_x_y_nist = self.weighted_cosine_correlation(a=0.5, b=1.3, nanfill=0)

        s_ss_x_y_nist = ((n_x * s_wc_x_y_nist) + (self.n_x_y * s_r_x_y)) / (n_x + self.n_x_y)    
        # final step

        return s_ss_x_y, s_ss_x_y_nist

    def pearson_correlation(self,):

        correlation = pearsonr(self.zero_filled_u_l[0], self.zero_filled_u_l[1])

        return correlation[0]

    def spearman_correlation(self):

        # calculate Spearman correlation
        # ## TODO - Check axis
        correlation = spearmanr(self.zero_filled_u_l[0], self.zero_filled_u_l[1], axis=0)

        return correlation[0]

    def kendall_tau(self):

        # create dict['mz'] = abundance, for experimental data
        # self.ms_mz_abun_dict = mass_spec.mz_abun_dict

        # create dict['mz'] = abundance, for experimental data

        # calculate Kendall's tau
        correlation = kendalltau(self.zero_filled_u_l[0], self.zero_filled_u_l[1])

        return correlation[0]

    def dft_correlation(self):

        if self.n_x_y == 0:
            return 0

        # count number of non-zero abundance/peak intensity values
        n_x = sum(a != 0 for a in self.exp_abun)

        x, y = self.nan_fill(self.df, fill_with=0)
        
        x, y = self.normalize(x, y, norm_func=self.normalize_func)
        
        # get the Fourier transform of x and y
        x_dft = rfft(x).real
        y_dft = rfft(y).real

        s_dft_xy = dot(x_dft, y_dft)/(norm(x_dft)*norm(y_dft))

        # using the existing weighted_cosine_correlation function to get S_WC(X,Y)
        s_wc_x_y = self.weighted_cosine_correlation(nanfill=0)

        # final step
        s_dft = (n_x * s_wc_x_y + self.n_x_y * s_dft_xy) / (n_x + self.n_x_y)

        return s_dft

    def dwt_correlation(self):

        if self.n_x_y == 0:
            return 0

        # count number of non-zero abundance/peak intensity values
        n_x = sum(a != 0 for a in self.exp_abun)

        # calculate cosine correlation,
        x, y = self.nan_fill(self.df, fill_with=0)
        
        x, y = self.normalize(x, y, norm_func=self.normalize_func)

        # Make x and y into an array
        x_a = list(x)
        y_a = list(y)

        # get the wavelet transform of x and y (Daubechies with a filter length of 4. Asymmetric. pywavelets function)
        # Will only use the detail dwt (dwtDd
        x_dwtD = dwt(x_a, 'db2')[1]
        y_dwtD = dwt(y_a, 'db2')[1]

        s_dwt_xy = dot(x_dwtD, y_dwtD) / (norm(x_dwtD) * norm(y_dwtD))

        # using the existing weighted_cosine_correlation function to get S_WC(X,Y)
        s_wc_x_y = self.weighted_cosine_correlation(nanfill=0)

        # final step
        s_dwt = (n_x * s_wc_x_y + self.n_x_y * s_dwt_xy) / (n_x + self.n_x_y)

        return s_dwt

    def euclidean_distance(self):

        # correlation = euclidean_distance_manual(self.zero_filled_u_l[0], self.zero_filled_u_l[1])
        qlist = self.zero_filled_u_l[0]
        rlist = self.zero_filled_u_l[1]

        correlation = sqrt(np_sum(power(qlist - rlist, 2)))

        return correlation

    def manhattan_distance(self):

        qlist = self.zero_filled_u_l[0]
        rlist = self.zero_filled_u_l[1]

        return np_sum(absolute(qlist - rlist))

    def jaccard_distance(self):

        def jaccard_similarity(list1, list2):

            intersection = len(list(set(list1).intersection(list2)))
            union = (len(list1) + len(list2)) - intersection
            return float(intersection) / union

        qlist = self.zero_filled_u_l[0]
        rlist = self.zero_filled_u_l[1]

        return np_sum(power(qlist - rlist, 2)) / (np_sum(power(qlist, 2)) + np_sum(power(rlist, 2)) - np_sum(qlist * rlist))
        # correlation = jaccard_similarity(self.zero_filled_u_l[0], self.zero_filled_u_l[1])
        # @return correlation

    def extra_distances(self):
        from corems.molecular_id.calc import math_distance

        #qlist = self.zero_filled_u_l[2]
        #rlist = self.zero_filled_u_l[3]

        dict_res = {}

        for method in methods_name:
            #function_name = method + "_distance"
            function_name = method
            if hasattr(math_distance, function_name):
                f = getattr(math_distance, function_name)

                if function_name == "canberra_metric":
                    
                    x, y = self.nan_fill(self.df, fill_with=0)
        
                    qlist, rlist = self.normalize(x, y, norm_func=self.normalize_func)
                    #print("qlist:")
                    #print(qlist)
                    #print("rlist:")
                    #print(rlist)

                
                else:
                    qlist = self.zero_filled_u_l[0]
                    rlist = self.zero_filled_u_l[1]

                dist = f(qlist, rlist)
                #if method == "Minokowski_3":
                #    print("qlist:")
                #    print(qlist)
                #    print("rlist")
                #    print(rlist)
                #    exit()
                # if dist == np.nan or dis == np.inf:
                    # print(self.exp_abun)
                    # print(self.exp_mz)
                    #print(function_name)
                    # print(len(self.exp_abun))
                    # print(len(self.exp_mz))
                    # print(self.zero_filled_u_l[1])
                dict_res[method] = dist

        return dict_res



import gc

from numpy import hamming, hanning, blackman, zeros, fft, sqrt, arange, where, power, absolute

__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"

class TransientCalculations(object):
    
    '''
    classdocs
    '''
    
    def cal_transient_time(self):
        
        return (1 / self.bandwidth) * ((self.number_data_points) / 2)
        
    def zero_fill(self, transient ):
        
        zeros_filled_transient = zeros(len(transient)*(self.parameters.number_of_zero_fills+1))
                
        zeros_filled_transient[0:len(transient)] = transient    
        
        del transient
        
        gc.collect()
        
        return  zeros_filled_transient 
    
    def truncation(self, transient):
        
        data_count = len(transient)
            
        for _ in range(self.parameters.number_of_truncations):
        
            data_count = int(data_count / 2)
         
        time_domain_truncated = transient[0:data_count]
        
        del transient
                
        gc.collect()  
        
        return time_domain_truncated
    
    def apodization(self, transient):
        
        apodi_method = self.parameters.apodization_method
        
        length = len(transient)
            
        if apodi_method == "Hamming":
                H_function = hamming(length)
        elif apodi_method == "Hanning":
                H_function = hanning(length)
        elif apodi_method == "Blackman":
                H_function = blackman(length)
            
        S_x = transient * H_function
        
        del transient
        gc.collect()  
            
        return S_x
    
    def calculate_frequency_domain(self, number_data_points):
        
        qntpoints = arange(0,(number_data_points))
        
        factor_distancy = (self.bandwidth)/(number_data_points)  
                
        frequency_domain = qntpoints * factor_distancy
        
        del qntpoints   
        del factor_distancy
        gc.collect()  
        
        return frequency_domain  
    
    def cut_freq_domain(self, freqdomain_X, freqdomain_Y):
      
        if self._exc_low_freq > self._exc_high_freq:
            
            final =  where(freqdomain_X > self._exc_high_freq)[-1][-1]
            comeco =  where(freqdomain_X > self._exc_high_freq())[0][0]
        
        else:
            
            final =  where(freqdomain_X > self._exc_low_freq)[-1][-1]
            comeco =  where(freqdomain_X > self._exc_low_freq)[0][0]
            
        
        return freqdomain_X[comeco:final], freqdomain_Y[comeco:final]
        #del freqdomain_X, freqdomain_Y
        #gc.collect()
    
    def phase_and_absorption_mode_ft(self):
        'anyone wants to play with this part please make yourself comfortable. I will:'
        pass 
            
    def perform_magniture_mode_ft(self, transient):
        
        A = fft.rfft(transient)
        
        #A = fft.fft(transient)
        #A = A[0:int(len(A)/2)]

        factor = int(self.parameters.number_of_zero_fills-1)
        if self.parameters.number_of_zero_fills:
            if self.parameters.number_of_zero_fills == 1:
                factor = 1/2
                
            else:
                factor = int(1/self.parameters.number_of_zero_fills+1)
                
            Max_index = int(len(A)/factor)    
        
        else:
            Max_index = int(len(A))
        
        A = A[0:Max_index]
        
        datapoints = len(A)
        
        freqdomain_X = self.calculate_frequency_domain(datapoints)
        
        magnitude_Y = sqrt((power(A.real,2)) + (power(A.imag,2)))
        
        freqdomain_X_cut, magnitude_Y_cut = self.cut_freq_domain(freqdomain_X, magnitude_Y)  
        
        del transient 
        #del freqdomain_X
        #del magnitude_Y
        gc.collect()
        
        return freqdomain_X_cut, magnitude_Y_cut
    
    def correct_dc_offset(self):
        
        pass
    
    





__author__ = "Yuri E. Corilo"
__date__ = "Feb 12, 2020"

from threading import Thread
from pathlib import Path

from corems.molecular_id.factory.EI_SQL import EI_LowRes_SQLite

class ReadNistMSI(Thread):

    def __init__(self, file_path, url='sqlite://'):
        
        Thread.__init__(self)
        file_path = Path(file_path)
        
        if not file_path.exists():
        
            raise FileExistsError("File does not exist: " + file_path)

        self.file_path = file_path

        self.url = url

    def run(self):
        
        self.sqlLite_obj = self.get_sqlLite_obj()

    def get_sqlLite_obj(self):

        compound_data_dict_list = self.get_compound_data_dict_list()

        sqlLite_obj = EI_LowRes_SQLite(url=self.url)
    
        for data_dict in compound_data_dict_list:
            if not data_dict.get('NUM PEAKS'):
                data_dict['NUM PEAKS'] = len(data_dict.get('mz'))
            if not data_dict.get('CASNO'):
                data_dict['CASNO'] = data_dict.get('CAS')
                if not data_dict['CASNO']:
                    data_dict['CASNO'] = 0
            #print(data_dict)
            try:
                sqlLite_obj.add_compound(data_dict)
            except:
                print(data_dict.get('NAME'))
        
        return sqlLite_obj

    def get_compound_data_dict_list(self,):
        
        list_dict_data = []
        
        with open(self.file_path) as msifile:
     
            content = msifile.readlines()
            
            i = 0
            
            dict_data = dict()
            dict_data['mz'] = list()
            dict_data['abundance'] = list()
            #for line in content:
            #   print(line, line=="\n" )
            
            while i < len(content):
                
                split_line = content[i].split(":") 
                
                #empty line 
                if len(content[i]) == 1:
                    
                    i += 1 
                    if dict_data.get('NAME'):
                        list_dict_data.append(dict_data)
                    
                    #print(dict_data)
                    dict_data = dict()
                    dict_data['mz'] = list()
                    dict_data['abundance'] = list()

                #metadata, name, ri, rt etc 
                elif len(split_line) >= 2:
                    
                    label = split_line[0]
                    value = ':'.join(split_line[1:]).strip('\n').strip('')
                    dict_data[label]= value
                    i += 1 
                
                #mz and abundance pairs
                elif len(split_line) == 1:
                    
                    for s in content[i].strip('\n').strip('').split("(")[1:]:
                    
                        values = s.split(" ") 
                        
                        if values[0] == '':
                            mz = values[1]
                        else:
                            mz = values[0]
                        
                        abun = values[-2].strip(')')

                        dict_data['mz'].append(mz)
                        dict_data['abundance'].append(abun)
                    
                    i += 1   
                #something else    
                else:
                    
                    i += 1    
        
        return list_dict_data

        


__author__ = "Yuri E. Corilo"
__date__ = "Jul 02, 2019"

from copy import deepcopy
import itertools
import multiprocessing
import json
import cProfile
import io
import pstats
import contextlib

from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import load_only
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine, func
from tqdm import tqdm

from corems.encapsulation.factory.processingSetting import MolecularLookupDictSettings
from corems.encapsulation.constant import Atoms
from corems.molecular_id.factory.molecularSQL import CarbonHydrogen, HeteroAtoms, MolecularFormulaLink
from corems.encapsulation.factory.parameters import MSParameters
from corems import chunks, timeit
from corems.molecular_id.factory.molecularSQL import MolForm_SQL
import os

@contextlib.contextmanager
def profiled():
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    print(s.getvalue())

def insert_database_worker(args):
        
        results, url = args
        
        if not url:
            
            url = 'sqlite:///db/molformulas.sqlite'

        if url[0:6] == 'sqlite':
            engine = create_engine(url, echo = False)
        else:
            engine = create_engine(url, echo = False, isolation_level="AUTOCOMMIT")
        
        session_factory = sessionmaker(bind=engine)
        session = session_factory()
        insert_query = MolecularFormulaLink.__table__.insert().values(results)
        session.execute(insert_query)
        session.commit()
        session.close()
        engine.dispose()

class MolecularCombinations:
     
    def __init__(self, sql_db = None):

        if not sql_db:
            
            self.sql_db = MolForm_SQL()
        else:
            
            self.sql_db = sql_db

    def cProfile_worker(self, args):
        
        cProfile.runctx('self.get_mol_formulas(*args)', globals(), locals(), 'mf_database_cprofile.prof')

    def check_database_get_class_list(self, molecular_search_settings):
        
        all_class_to_create = []
        
        classes_dict = self.get_classes_in_order(molecular_search_settings)
        
        class_str_set = set(classes_dict.keys())
        
        existing_classes_objs = self.sql_db.session.query(HeteroAtoms).distinct().all()
        
        existing_classes_str = set([classe.name for classe in existing_classes_objs])

        self.len_existing_classes = len(existing_classes_str)

        class_to_create = class_str_set - existing_classes_str
        
        class_count= len(existing_classes_objs)

        data_classes = list()    
        for index, class_str in enumerate(class_to_create):
            
            class_dict = classes_dict.get(class_str)
            halogen_count = self.get_total_halogen_atoms(class_dict)
            data_classes.append({"name":class_str, "id":class_count+ index + 1, "halogensCount": halogen_count})

        #data_classes = [{"name":class_str, "id":class_count+ index + 1} for index, class_str in enumerate(class_to_create)]

        if data_classes:

            list_insert_chunks = chunks(data_classes, self.sql_db.chunks_count)
            for insert_chunk in  list_insert_chunks:   
                insert_query = HeteroAtoms.__table__.insert().values(insert_chunk)
                self.sql_db.session.execute(insert_query)

        for index, class_str in enumerate(class_to_create):

            class_tuple = (class_str, classes_dict.get(class_str), class_count+ index + 1)

            all_class_to_create.append(class_tuple)

        return [(c_s, c_d) for c_s, c_d in classes_dict.items()], all_class_to_create, existing_classes_objs       

    def get_carbonsHydrogens(self, settings, odd_even):

        operator = '==' if odd_even == 'even' else '!=' 
        usedAtoms = settings.usedAtoms
        user_min_c, user_max_c = usedAtoms.get('C')
        user_min_h, user_max_h = usedAtoms.get('H')

        return eval("self.sql_db.session.query(CarbonHydrogen).filter(" 
                                       "CarbonHydrogen.C >= user_min_c,"
                                        "CarbonHydrogen.H >= user_min_h,"
                                        "CarbonHydrogen.C <= user_max_c,"
                                        "CarbonHydrogen.H <= user_max_h,"
                                        "CarbonHydrogen.H % 2" + operator+ "0).all()")

    def add_carbonsHydrogens(self, settings, existing_classes_objs):

        usedAtoms = settings.usedAtoms

        user_min_c, user_max_c = usedAtoms.get('C')
        user_min_h, user_max_h = usedAtoms.get('H')

        query_obj = self.sql_db.session.query(func.max(CarbonHydrogen.C).label("max_c"), 
                        func.min(CarbonHydrogen.C).label("min_c"),
                        func.max(CarbonHydrogen.H).label("max_h"),
                        func.min(CarbonHydrogen.H).label("min_h"),
                        )


        database = query_obj.first()
        if database.max_c == user_max_c and database.min_c == user_min_c and database.max_h == user_max_h and database.min_h == user_min_h:   
            #all data is already available at the database
            pass

        else:

            current_count = self.sql_db.session.query(CarbonHydrogen.C).count()
            
            databaseCarbonHydrogen = self.sql_db.session.query(CarbonHydrogen).all()
            
            userCarbon = set(range(user_min_c, user_max_c + 1))
            userHydrogen = set(range(user_min_h, user_max_h + 1))
            
            carbon_hydrogen_objs_database = {}
            for obj in databaseCarbonHydrogen:
                
                str_data = "C:{},H:{}".format(obj.C, obj.H)
                carbon_hydrogen_objs_database[str_data] = str_data

            carbon_hydrogen_objs_to_create = {'even': {}, 'odd': {}}
            
            list_ch_obj_to_add = list()
            i = 0
            for comb in itertools.product(userCarbon, userHydrogen):
                
                C  = comb[0]
                H =  comb[1]
                data = {"C":C,
                       "H":H,
                }

                data_insert = {"C":C,
                       "H":H,
                }

                str_data = "C:{},H:{}".format(C,H)
                
                if not str_data in carbon_hydrogen_objs_database.keys():
                    
                    label = 'even' if comb[1]%2 == 0 else 'odd'
                    data["mass"] = (C * Atoms.atomic_masses.get('C')) + (H * Atoms.atomic_masses.get('H'))
                    data["dbe"] = C - (H/2) + 1
                    data["id"] = i + current_count + 1
                    data_insert["id"] = i + current_count + 1
                    i = i + 1 
                    carbon_hydrogen_objs_to_create[label][str_data] = data
                    
                    list_ch_obj_to_add.append(data_insert)

            if list_ch_obj_to_add:
                # insert carbon hydrogen objs
                list_insert_chunks = chunks(list_ch_obj_to_add, self.sql_db.chunks_count)
                for insert_chunk in  list_insert_chunks:   
                    insert_query = CarbonHydrogen.__table__.insert().values(insert_chunk)
                    self.sql_db.session.execute(insert_query)
                self.sql_db.session.commit()    
            
                
                list_molecular_form= list()
                for classe_obj in existing_classes_objs:

                    classe_dict = classe_obj.to_dict()  
                    classe_mass = self.calc_mz(classe_dict)
                    classe_dbe = self.calc_dbe_class(classe_dict)

                    odd_even_label = self.get_h_odd_or_even(classe_dict)

                    ch_datalist = carbon_hydrogen_objs_to_create.get(odd_even_label).values()

                    for ch_dict in ch_datalist:
                        mass = ch_dict.get('mass') + classe_mass
                        dbe = ch_dict.get('dbe') + classe_dbe

                        if settings.min_mz <= mass <= settings.max_mz:
                
                            if settings.min_dbe <= dbe <= settings.max_dbe:
                                
                                list_molecular_form.append( {"heteroAtoms_id":classe_obj.id, 
                                        "carbonHydrogen_id":ch_dict.get('id'), 
                                        "mass":mass, "DBE":dbe})

                list_insert_chunks = chunks(list_molecular_form, self.sql_db.chunks_count)
                for insert_chunk in  list_insert_chunks:   
                    insert_query = MolecularFormulaLink.__table__.insert().values(insert_chunk)
                    self.sql_db.session.execute(insert_query)
                self.sql_db.session.commit()    
            
                
    @timeit
    def runworker(self, molecular_search_settings):
        
        classes_list, class_to_create, existing_classes_objs = self.check_database_get_class_list(molecular_search_settings)
        
        settings = MolecularLookupDictSettings()
        settings.usedAtoms = deepcopy(molecular_search_settings.usedAtoms)
        settings.url_database = molecular_search_settings.url_database
        settings.db_jobs = molecular_search_settings.db_jobs

        self.add_carbonsHydrogens(settings, existing_classes_objs)
        
        if class_to_create:
            
            settings = MolecularLookupDictSettings()
            settings.usedAtoms = deepcopy(molecular_search_settings.usedAtoms)
            settings.url_database = molecular_search_settings.url_database
            settings.db_jobs = molecular_search_settings.db_jobs
            
            self.sql_db.session.commit()
            odd_ch_obj = self.get_carbonsHydrogens(settings,'odd')
            self.odd_ch_id = [obj.id for obj in odd_ch_obj]
            self.odd_ch_dict = [{'C':obj.C, 'H':obj.H} for obj in odd_ch_obj]
            self.odd_ch_mass = [obj.mass for obj in odd_ch_obj]
            self.odd_ch_dbe = [obj.dbe for obj in odd_ch_obj]
            
            even_ch_obj = self.get_carbonsHydrogens(settings, 'even')
            self.even_ch_id = [obj.id for obj in even_ch_obj]
            self.even_ch_dict = [{'C':obj.C, 'H':obj.H} for obj in even_ch_obj]
            self.even_ch_mass = [obj.mass for obj in even_ch_obj]
            self.even_ch_dbe = [obj.dbe for obj in even_ch_obj]

            all_results= list()
            for class_tuple in tqdm(class_to_create):
                
                results = self.populate_combinations(class_tuple, settings)
                all_results.extend(results)
                if settings.db_jobs == 1: 
                    #if len(all_results) >= self.sql_db.chunks_count:
                        list_insert_chunks = list(chunks(results, self.sql_db.chunks_count))
                        for chunk in list_insert_chunks:
                            insert_query = MolecularFormulaLink.__table__.insert().values(chunk)
                            self.sql_db.session.execute(insert_query)
                        #all_results = list()
            self.sql_db.session.commit()
            # each chunk takes ~600Mb of memory, so if using 8 processes the total free memory needs to be 5GB
            if settings.db_jobs > 1: 
                list_insert_chunks = list(chunks(all_results, self.sql_db.chunks_count))
                print( "Started database insert using {} iterations for a total of {} rows".format(len(list_insert_chunks), len(all_results)))
                worker_args = [(chunk, settings.url_database) for chunk in list_insert_chunks]
                p = multiprocessing.Pool(settings.db_jobs)
                for class_list in tqdm(p.imap_unordered(insert_database_worker, worker_args)):
                    pass
                p.close()
                p.join()
        
        return classes_list
    
    def get_classes_in_order(self, molecular_search_settings):
        ''' structure is 
            ('HC', {'HC': 1})'''
        
        usedAtoms = deepcopy(molecular_search_settings.usedAtoms)
        
        usedAtoms.pop("C")
        usedAtoms.pop("H")

        min_n, max_n = usedAtoms.get('N') if usedAtoms.get('N') else (0,0)
        min_o, max_o = usedAtoms.get('O') if usedAtoms.get('O') else (0,0)
        min_s, max_s = usedAtoms.get('S') if usedAtoms.get('S') else (0,0)
        min_p, max_p = usedAtoms.get('P') if usedAtoms.get('P') else (0,0)

        possible_n = [n for n in range(min_n, max_n + 1)]
        possible_o = [o for o in range(min_o, max_o + 1)]
        possible_s = [s for s in range(min_s, max_s + 1)]
        possible_p = [p for p in range(min_p, max_p + 1)]
        
        atoms_in_order = ['N', 'O', 'S', 'P']

        classe_in_order = {}

        all_atoms_tuples = itertools.product(possible_n, possible_o,
                                            possible_s, possible_p)
        
        for atom in atoms_in_order:
            usedAtoms.pop(atom, None)
        
        for selected_atom, min_max_tuple in usedAtoms.items():
            
            min_x = min_max_tuple[0]
            max_x = min_max_tuple[1]
            

            possible_x = [x for x in range(min_x, max_x + 1)]

            all_atoms_tuples = itertools.product(all_atoms_tuples, possible_x)
            all_atoms_tuples = [all_atoms_combined[0] + (all_atoms_combined[1],) for all_atoms_combined in
                                all_atoms_tuples]
            atoms_in_order.append(selected_atom)
        
        for all_atoms_tuple in all_atoms_tuples:

            classe_str = ''
            classe_dict = {}
            
            for each_atoms_index, atom_number in enumerate(all_atoms_tuple):
                
                if atom_number != 0:
                    classe_dict[atoms_in_order[each_atoms_index]] = atom_number
            
            if not classe_dict:
                classe_in_order['HC'] = {"HC": ""}
                continue

            classe_str =json.dumps(classe_dict)
            
            if len(classe_str) > 0:
                
                classe_in_order[classe_str] =  classe_dict
        
        classe_in_order_dict = self.sort_classes(atoms_in_order, classe_in_order)
        
        return classe_in_order_dict

    @staticmethod
    def sort_classes( atoms_in_order, combination_dict) -> [str]: 
        #ensures atoms are always in the order defined at atoms_in_order list
        join_dict_classes = dict()
        atoms_in_order =  ['N','S','P','O'] + atoms_in_order[4:] + ['HC']
        
        sort_method = lambda atoms_keys: [atoms_in_order.index(atoms_keys)] 
        for class_str, class_dict in combination_dict.items():
            
            sorted_dict_keys = sorted(class_dict, key = sort_method)
            class_dict = { atom: class_dict[atom] for atom in sorted_dict_keys}
            class_str = json.dumps(class_dict)
            # using json for the new database, class 
            # class_str = ' '.join([atom + str(class_dict[atom]) for atom in sorted_dict_keys])
            join_dict_classes[class_str] =  class_dict
        
        return join_dict_classes

    @staticmethod
    def get_fixed_initial_number_of_hydrogen( min_h, odd_even):

        remaining_h = min_h % 2
        
        if odd_even == 'even':
            
            if remaining_h == 0: return remaining_h
            
            else: return remaining_h + 1    
        
        else:
            
            if remaining_h == 0: return remaining_h + 1
            
            else: return remaining_h    

    def calc_mz(self, datadict, class_mass=0):
        
        mass = class_mass
        
        for atom in datadict.keys():
            
            if atom != 'HC':    
                
                mass = mass + Atoms.atomic_masses[atom]  *  datadict.get(atom)
            
        return mass 
        
    def calc_dbe_class(self, datadict):
            
            init_dbe = 0
            for atom in datadict.keys():

                if atom == 'HC':
                    continue  
                
                n_atom = int(datadict.get(atom))
                
                clean_atom = ''.join([i for i in atom if not i.isdigit()]) 
                
                valencia = MSParameters.molecular_search.used_atom_valences.get(clean_atom)
                
                if type(valencia) is tuple:
                    valencia = valencia[0]
                if valencia > 0:
                    #print atom, valencia, n_atom, init_dbe
                    init_dbe = init_dbe + (n_atom * (valencia - 2))
                else:
                    continue
                
            return (0.5 * init_dbe)
            
    def populate_combinations(self, classe_tuple, settings):
        
        ion_charge =  0
        
        class_dict = classe_tuple[1]
        odd_or_even = self.get_h_odd_or_even(class_dict)
        
        return self.get_mol_formulas(odd_or_even, classe_tuple, settings)
        
    def get_or_add(self, SomeClass, kw):
            
        obj = self.sql_db.session.query(SomeClass).filter_by(**kw).first()
        if not obj:
            obj = SomeClass(**kw)
        return obj
    
    
    def get_mol_formulas(self, odd_even_tag, classe_tuple, settings):
        
        class_str = classe_tuple[0]
        class_dict = classe_tuple[1]
        classe_id = classe_tuple[2]
        
        results = list()
        
        if 'HC' in class_dict:
            del class_dict['HC']
            
        class_dbe = self.calc_dbe_class(class_dict)    
        class_mass = self.calc_mz(class_dict)
        
        carbonHydrogen_mass = self.odd_ch_mass if odd_even_tag == 'odd' else self.even_ch_mass 
        carbonHydrogen_dbe = self.odd_ch_dbe if odd_even_tag == 'odd' else self.even_ch_dbe 
        carbonHydrogen_id = self.odd_ch_id if odd_even_tag == 'odd' else self.even_ch_id 
        
        for index, carbonHydrogen_obj in enumerate(carbonHydrogen_id):
            
            mass = carbonHydrogen_mass[index] + class_mass
            dbe =  carbonHydrogen_dbe[index] + class_dbe
    
            if settings.min_mz <= mass <= settings.max_mz:
                
                if settings.min_dbe <= dbe <= settings.max_dbe:
                    
                    molecularFormula=  {"heteroAtoms_id":classe_id, 
                            "carbonHydrogen_id":carbonHydrogen_id[index], 
                            "mass":mass, "DBE":dbe}
                    
                    results.append(molecularFormula)
        
        return results
        
        
    def get_h_odd_or_even(self, class_dict):

        
        HAS_NITROGEN = 'N' in class_dict.keys()
        HAS_PHOSPHORUS = 'P' in class_dict.keys()

        number_of_halogen = self.get_total_halogen_atoms(class_dict)

        if number_of_halogen > 0:

            TEM_HALOGEN = True

        else:

            TEM_HALOGEN = False

        if TEM_HALOGEN:

            remaining_halogen = number_of_halogen % 2

        else:

            remaining_halogen = 0

        if HAS_NITROGEN and HAS_PHOSPHORUS:

            number_of_n = class_dict.get('N') + class_dict.get('P')
            remaining_n = number_of_n % 2

        elif HAS_NITROGEN and not HAS_PHOSPHORUS:

            number_of_n = class_dict.get('N')
            remaining_n = number_of_n % 2

        elif HAS_PHOSPHORUS and not HAS_NITROGEN:

            number_of_n = class_dict.get('P')
            remaining_n = number_of_n % 2

        else:

            remaining_n = -1

        if remaining_n > 0.0:
            if HAS_NITROGEN or HAS_PHOSPHORUS:

                if TEM_HALOGEN:
                    if remaining_halogen == 0:
                        return 'odd'
                    else:
                        return 'even'
                else:
                    return 'odd'

        elif remaining_n == 0.0:

            if HAS_NITROGEN or HAS_PHOSPHORUS:

                if TEM_HALOGEN:
                    if remaining_halogen == 0:
                        return 'even'
                    else:
                        return 'odd'
                else:
                    return 'even'

        else:

            if TEM_HALOGEN:
                if remaining_halogen == 0:
                    return 'even'
                else:
                    return 'odd'
            else:
                return 'even'

    @staticmethod
    def get_total_halogen_atoms(class_dict):

            atoms = ['F', 'Cl', 'Br', 'I']

            total_number = 0
            
            for atom in atoms:

                if atom in class_dict.keys():

                    total_number = total_number + class_dict.get(atom)
            
            return total_number    



"""
Created on Jun 12, 2019
"""
from collections.abc import Mapping
from pathlib import Path
from corems.mass_spectra.calc.LC_Calc import LC_Calculations

__author__ = "Yuri E. Corilo"
__date__ = "Jun 25, 2019"

class LCMSBase(Mapping, LC_Calculations):
    """
    classdocs
    """

    def __init__(self, file_location, analyzer='Unknown', instrument_label='Unknown', sample_name=None):
        
        '''
         # Parameters
		----------
		file_location: text,  pathlib.Path(), or s3path.S3Path 
            Path object from pathlib containing the file location
        '''
		
        if  isinstance(file_location, str):
			# if obj is a string it defaults to create a Path obj, pass the S3Path if needed
            file_location = Path(file_location)

        if not file_location.exists():
        
            raise FileExistsError("File does not exist: " + file_location)
        
        self.file_location = file_location
        
        if sample_name: 
        
            self.sample_name = sample_name

        else: 
        
            self.sample_name = file_location.stem
        
        self.analyzer = analyzer
        self.instrument_label = instrument_label
        
        self._retention_time_list = []
        self._scans_number_list = []
        self._tic_list = []

        self._ms = {}
        """
        key is scan number; value is MassSpectrum Class
        """
    
    def __len__(self):
        
        return len(self._ms)
        
    def __getitem__(self, scan_number):
        
        return self._ms.get(scan_number)

    def __iter__(self):

         return iter(self._ms.values()) 

    def add_mass_spectrum(self, mass_spec):

        self._ms[mass_spec.scan_number] = mass_spec

    def set_tic_list_from_data(self):

        self.tic = [self._ms.get(i).tic for i in self.scans_number]
        
        # self.set_tic_list([self._ms.get(i).get_sumed_signal_to_noise() for i in self.get_scans_number()])

    def set_retention_time_from_data(self):

        retention_time_list = []

        for key_ms in sorted(self._ms.keys()):

            retention_time_list.append(self._ms.get(key_ms).rt)

        self.retention_time = retention_time_list 

        # self.set_retention_time_list(sorted(self._ms.keys()))

    def set_scans_number_from_data(self):
        
        self.scans_number = sorted(self._ms.keys())

    @property
    def scans_number(self):

        return self._scans_number_list

    @property
    def retention_time(self):

        return self._retention_time_list
    
    @property
    def tic(self):

        return self._tic_list

    @retention_time.setter
    def retention_time(self, l):
        # self._retention_time_list = linspace(0, 80, num=len(self._scans_number_list))
        self._retention_time_list = l

    @scans_number.setter
    def scans_number(self, l):

        self._scans_number_list = l

    @tic.setter
    def tic(self, l):

        self._tic_list = l    

    


import sys
sys.path.append(".")
import os

from sqlalchemy import create_engine, ForeignKey, Column, Integer, String, Float, SMALLINT
from sqlalchemy.orm import backref, column_property, relationship
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.sql.schema import UniqueConstraint
from sqlalchemy import exc

from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.sql.operators import exists
from sqlalchemy import event, and_
from sqlalchemy import func

from corems.encapsulation.constant import Atoms, Labels
import json
from corems.encapsulation.factory.processingSetting import MolecularFormulaSearchSettings
from sqlalchemy.orm.scoping import scoped_session
from corems import chunks
import tqdm

Base = declarative_base()

class HeteroAtoms(Base):

    __tablename__ = 'heteroAtoms'

    id = Column(Integer, primary_key=True,
                unique=True,
                nullable=False)

    name = Column(String, unique=True, nullable=False)

    halogensCount = Column(Integer, unique=False, nullable=False)

    carbonHydrogen = relationship('CarbonHydrogen', secondary='molecularformula')

    def __repr__(self):
        return '<HeteroAtoms Model {} class {}>'.format(self.id, self.name)      

    @hybrid_property
    def halogens_count(cls):
        return cls.halogensCount.cast(Float)

    def to_dict(self):

        return json.loads(self.name)


class CarbonHydrogen(Base):

    __tablename__ = 'carbonHydrogen'
    __table_args__ = (UniqueConstraint('C', 'H', name='unique_c_h'), )

    id = Column(Integer, primary_key=True,
                unique=True,
                nullable=False)

    C = Column(Integer, nullable=False)

    H = Column(Integer, nullable=False)

    heteroAtoms = relationship("HeteroAtoms",
                               secondary="molecularformula",
                               )

    def __repr__(self):
        return '<CarbonHydrogen Model {} C{} H{}>'.format(self.id, self.C, self.H)                     

    @property
    def mass(self):
        return (self.C * Atoms.atomic_masses.get('C')) + (self.H * Atoms.atomic_masses.get('H'))

    @hybrid_property
    def c(cls):
        return cls.C.cast(Float)

    @hybrid_property
    def h(cls):
        return cls.H.cast(Float)

    @hybrid_property
    def dbe(cls):
        # return cls.C.cast(Float) - (cls.H.cast(Float) / 2) + 1
        return float(cls.C) - float(cls.H / 2) + 1

# 264888.88 ms
class MolecularFormulaLink(Base):

    __tablename__ = 'molecularformula'
    __table_args__ = (UniqueConstraint('heteroAtoms_id', 'carbonHydrogen_id', name='unique_molform'), )

    # id = Column(Integer, primary_key=True,
    #                    unique=True,
    #                    nullable=False)

    heteroAtoms_id = Column(Integer,
                            ForeignKey('heteroAtoms.id'),
                            primary_key=True)

    carbonHydrogen_id = Column(Integer,
                               ForeignKey('carbonHydrogen.id'), 
                               primary_key=True)

    mass = Column(Float)

    DBE = Column(Float)

    carbonHydrogen = relationship(CarbonHydrogen, backref=backref("heteroAtoms_assoc"))

    heteroAtoms = relationship(HeteroAtoms, backref=backref("carbonHydrogen_assoc"))

    C = association_proxy('carbonHydrogen', 'C')

    H = association_proxy('carbonHydrogen', 'H')

    classe = association_proxy('heteroAtoms', 'name')

    @property
    def formula_dict(self):

        carbon = {'C': self.C, 'H': self.H}
        classe = json.loads(self.classe)
        if self.classe == '{"HC": ""}':
            return {**carbon}
        else:
            return {**carbon, **classe}

    @property
    def formula_string(self):
        class_dict = self.formula_dict
        class_str = ' '.join([atom + str(class_dict[atom]) for atom in class_dict.keys()])
        return class_str.strip()

    @property
    def classe_string(self):
        class_dict = json.loads(self.classe)
        class_str = ' '.join([atom + str(class_dict[atom]) for atom in class_dict.keys()])
        return class_str.strip()

    @hybrid_method
    def adduct_mass(self, ion_charge, adduct_atom):
        return (self.mass + (Atoms.atomic_masses.get(adduct_atom)) + (ion_charge * -1 * Atoms.electron_mass))/ abs(ion_charge)

    @hybrid_method
    def protonated_mass(self, ion_charge):
        return (self.mass + (ion_charge * Atoms.atomic_masses.get("H")) + (ion_charge * -1 * Atoms.electron_mass))/abs(ion_charge)

    @hybrid_method
    def radical_mass(self, ion_charge):
        return (self.mass + (ion_charge * -1 * Atoms.electron_mass))/ abs(ion_charge)

    def __repr__(self):

        return '<MolecularFormulaLink Model {}>'.format(self.formula_string)       


class MolForm_SQL:

    def __init__(self, url=None, echo=False):

        self.engine = self.init_engine(url)

        self.add_engine_pidguard(self.engine)

        session_factory = sessionmaker(bind=self.engine)

        Session = scoped_session(session_factory)

        self.session = session_factory()

        Base.metadata.create_all(self.engine)

        self.session.commit()

    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the dbconnection gets closed

        self.commit()
        self.session.close()
        self.engine.dispose()

    def initiate_database(self, url, database_name):  #CREATION

        engine = create_engine(url)
        conn = engine.connect()
        conn.execute("commit")
        conn.execute("create database " + database_name)
        conn.close()

    def commit(self):

        try:
            self.session.commit()  
        except SQLAlchemyError as e:
            self.session.rollback()
            print(str(e))

    def init_engine(self, url):

        if not url or url == 'None' or url == 'False':
            directory = os.getcwd()

            if not os.path.isdir(directory + '/db'):
                os.mkdir(directory + '/db')

            url = 'sqlite:///{DB}/db/molformulas.sqlite'.format(DB=directory)

        if url[0:6] == 'sqlite':
            self.type = 'sqlite'
        else:
            self.type = 'normal'
            
        if url[0:6] == 'sqlite':
            engine = create_engine(url, echo = False)
            self.chunks_count = 50
        
        elif url[0:10] == 'postgresql' or url[0:8] == 'postgres':
            #postgresql
            self.chunks_count = 50000
            engine = create_engine(url, echo = False, isolation_level="AUTOCOMMIT")
        
        return engine# poolclass=NullPool

    def __enter__(self):
        
        return self
    
    def get_dict_by_classes(self, classes, ion_type, nominal_mzs, ion_charge, molecular_search_settings, adducts=None):
                                    
        '''Known issue, when using SQLite:
         if the number of classes and nominal_m/zs are higher than 999 the query will fail
         Solution: use postgres or split query''' 
        def query_normal(class_list, len_adduct):
            
            base_query = self.session.query(MolecularFormulaLink, CarbonHydrogen, HeteroAtoms)\
                                .filter(MolecularFormulaLink.carbonHydrogen_id == CarbonHydrogen.id)\
                                .filter(MolecularFormulaLink.heteroAtoms_id == HeteroAtoms.id)
            
            return base_query.filter(
                and_(
                    HeteroAtoms.name.in_(class_list), 
                    and_(
                        MolecularFormulaLink.DBE >= molecular_search_settings.min_dbe, 
                        MolecularFormulaLink.DBE <= molecular_search_settings.max_dbe, 
                        and_(
                            ((CarbonHydrogen.h + HeteroAtoms.halogens_count - len_adduct) / CarbonHydrogen.c) >= molecular_search_settings.min_hc_filter,
                            ((CarbonHydrogen.h + HeteroAtoms.halogens_count - len_adduct) / CarbonHydrogen.c) <= molecular_search_settings.max_hc_filter,
                            CarbonHydrogen.C >= molecular_search_settings.usedAtoms.get("C")[0],
                            CarbonHydrogen.c <= molecular_search_settings.usedAtoms.get("C")[1],
                            CarbonHydrogen.h >= molecular_search_settings.usedAtoms.get("H")[0],
                            CarbonHydrogen.h <= molecular_search_settings.usedAtoms.get("H")[1],
                        )
                    )
                )
            )

        def add_dict_formula(formulas, ion_type, ion_charge, adduct_atom=None):
            "organize data by heteroatom classes"
            dict_res = {}

            def nominal_mass_by_ion_type(formula_obj):
                
                if ion_type == Labels.protonated_de_ion:
                
                    return int(formula_obj.protonated_mass(ion_charge))
                
                elif ion_type == Labels.radical_ion:
                    
                    return int(formula_obj.radical_mass(ion_charge))

                elif ion_type == Labels.adduct_ion and adduct_atom:
                    
                    return int(formula_obj.adduct_mass(ion_charge, adduct_atom))
            
            for formula_obj, ch_obj, classe_obj in tqdm.tqdm(formulas, desc="Loading molecular formula database"):
                
                nominal_mz = nominal_mass_by_ion_type(formula_obj)
                
                if self.type != 'normal':
                    if not nominal_mz in nominal_mzs:
                        continue
                classe = classe_obj.name

                # classe_str = formula.classe_string
                
                # pbar.set_description_str(desc="Loading molecular formula database for class %s " % classe_str)
                
                formula_dict = formula_obj.formula_dict

                if formula_dict.get("O"):
                    
                    if formula_dict.get("O") / formula_dict.get("C") >= molecular_search_settings.max_oc_filter:
                        # print(formula_dict.get("O") / formula_dict.get("C"), molecular_search_settings.max_oc_filter)
                        continue
                    elif formula_dict.get("O") / formula_dict.get("C") <= molecular_search_settings.min_oc_filter:
                        # print(formula_dict.get("O") / formula_dict.get("C"), molecular_search_settings.min_oc_filter)
                        continue
                    #if formula_dict.get("P"):

                    #    if  not (formula_dict.get("O") -2)/ formula_dict.get("P") >= molecular_search_settings.min_op_filter:
                            
                    #        continue
        
                if classe in dict_res.keys():
                    
                    if nominal_mz in dict_res[classe].keys():
                        
                        dict_res.get(classe).get(nominal_mz).append(formula_obj)
                    
                    else:

                        dict_res.get(classe)[nominal_mz] = [formula_obj ]  
            
                else:
                    
                    dict_res[classe] = {nominal_mz: [formula_obj] }     
            
            return dict_res
        
        
        len_adducts = 0
        if ion_type == Labels.adduct_ion:
            len_adducts = 1
        
        query = query_normal(classes, len_adducts)

        if ion_type == Labels.protonated_de_ion:
            if self.type == 'normal':
                query = query.filter(and_(
                                MolecularFormulaLink.protonated_mass(ion_charge).cast(Integer).in_(nominal_mzs)
                                ))
            return add_dict_formula(query, ion_type, ion_charge)
        
        if ion_type == Labels.radical_ion:
            if self.type == 'normal':
                query = query.filter(MolecularFormulaLink.radical_mass(ion_charge).cast(Integer).in_(nominal_mzs))    
            return add_dict_formula(query, ion_type, ion_charge)
        
        if ion_type == Labels.adduct_ion:
            dict_res = {}
            if adducts: 
                for atom in adducts:
                    if self.type == 'normal':
                        query = query.filter(MolecularFormulaLink.adduct_mass(ion_charge, atom).cast(Integer).in_(nominal_mzs))    
                    dict_res[atom] = add_dict_formula(query, ion_type, ion_charge, adduct_atom=atom)
                return dict_res
        # dump all objs to memory
        self.session.expunge_all()
        
    def check_entry(self,classe, ion_type, molecular_search_settings):
        #  get all classes, ion_type, ion charge as str add to a dict or list
        #  then check if class in database
        has_class = self.session.query(exists().where(
            (MolecularFormulaLink.classe == classe)))
        
        return has_class
    
    def get_all_classes(self):
        
        query = self.session.query(MolecularFormulaLink.classe.distinct().label("classe"))
        
        return query.all()  
    
    def get_all(self,):
        
        mol_formulas = self.session.query(MolecularFormulaLink).all()
        
        return mol_formulas

    def delete_entry(self, row):
        
        try:
            self.session.delete(row)  
            self.session.commit()  
        
        except SQLAlchemyError as e:
            self.session.rollback()
            print(str(e))

    def purge(self, cls):
        '''Carefull, this will delete the entire database table'''
        self.session.query(cls).delete()
        self.session.commit()  

    def clear_data(self):
        '''clear tables'''
        meta = Base.metadata
        for table in reversed(meta.sorted_tables):
            print ('Clear table %s' % table)
            self.session.execute(table.delete())
        self.session.commit()

    def close(self, commit=True):
        # make sure the dbconnection gets closed
        
        if commit: self.commit()
        self.session.close()
        self.engine.dispose()    
   
    def add_engine_pidguard(self, engine):
        import os, warnings
        """Add multiprocessing guards.

        Forces a connection to be reconnected if it is detected
        as having been shared to a sub-process.

        """

        @event.listens_for(engine, "connect")
        def connect(dbapi_connection, connection_record):
            connection_record.info['pid'] = os.getpid()

        @event.listens_for(engine, "checkout")
        def checkout(dbapi_connection, connection_record, connection_proxy):
            pid = os.getpid()
            if connection_record.info['pid'] != pid:
                # substitute log.debug() or similar here as desired
                warnings.warn(
                    "Parent process %(orig)s forked (%(newproc)s) with an open "
                    "database connection, "
                    "which is being discarded and recreated." %
                    {"newproc": pid, "orig": connection_record.info['pid']})
                connection_record.connection = connection_proxy.connection = None
                raise exc.DisconnectionError(
                    "Connection record belongs to pid %s, "
                    "attempting to check out in pid %s" %
                    (connection_record.info['pid'], pid)
                )    

if __name__ == "__main__":
    
    sql = MolForm_SQL(url='sqlite:///')
    
    dict_data = {"name": '{"O": 12}'}
    dict_data2 = {"name": '{"O": 13}'}
    hetero_obj = HeteroAtoms(**dict_data)
    hetero_obj2 = HeteroAtoms(**dict_data2)
    sql.session.add(hetero_obj)
    sql.session.add(hetero_obj2)

    print(sql.session.query(HeteroAtoms).all())
    #molecular_search_settings = MolecularFormulaSearchSettings()
    #sql = MolForm_SQL()
    #query = sql.session.query(MolecularFormulaLink).filter_by(classe = '{"O": 12}').filter(MolecularFormulaLink.adduct_mass(+2, "Na") < 250)
    #query = sql.get_by_classe('{"O": 12}', molecular_search_settings).filter(MolecularFormulaLink.adduct_mass(+2, "Na") < 250)
    #classes = ['{"O": 12}']*1
    #for i, classe in enumerate(classes):
        #query = sql.get_by_classe(classe, molecular_search_settings)
        #query = sql.session.query(MolecularFormulaLink).filter_by(classe = '{"O": 12}').filter(MolecularFormulaLink.adduct_mass(+2, "Na") < 250)
        #for i in query.filter(MolecularFormulaLink.mass < 250):
            
        #    print(i.radical_mass(-1), i.protonated_mass(-1), i.adduct_mass(+2, "Na"), i.mass, i.formula_dict, i.formula_string)
    #
 


import os
import pathlib
import sys
from setuptools import setup, find_packages
# The directory containing this file

# The text of the README file
with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

# This call to setup() does all the work
setup(
    name="CoreMS",
    version="1.1.10",
    description="Mass Spectrometry Framework for Small Molecules Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EMSL-Computing/CoreMS",
    author="Corilo, Yuri",
    author_email="corilo@pnnl.gov",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 4 - Beta"
    ],


    package_data={'external': ['disclaimer.txt'], '': ['ext_lib/*']},
    packages=find_packages(),
    exclude_package_data={'.': ["tests", "*.win_only"]},
    include_package_data=True,
    install_requires=required,
    setup_requires=['pytest-runner', 'wheel'],
    test_suite='pytest',
    tests_require=['pytest'],
)



# -*- coding: utf-8 -*-
"""
Created on Wed May 13 02:16:09 2020

@author: Will Kew
"""

### import modules
import pandas as pd
import numpy as np
import os
import csv
from io import BytesIO
from pathlib import Path

from s3path import S3Path
# import corems modules
from corems.transient.input.brukerSolarix import ReadBrukerSolarix
from corems.encapsulation.factory.parameters import MSParameters
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas

# import scipy modules for calibration
from scipy.optimize import minimize

class MzDomainCalibration:

    def __init__(self, mass_spectrum, ref_masslist):
        
        self.mass_spectrum = mass_spectrum

        # define reference mass list - bruker .ref format
        self.ref_mass_list_path = ref_masslist

        self.mass_spectrum.mz_cal = self.mass_spectrum.mz_exp    
    
        print("MS Obj loaded - "+str(len(mass_spectrum.mspeaks))+" peaks found.")


    def load_ref_mass_list(self, refmasslist):
        """
        function to load in bruker mass list format into a dataframe


        Parameters
        ----------
        refmasslist : str
            full path to reference mass list (.ref) to import.

        Returns
        -------
        df_ref : pandas dataframe
            reference mass list object.

        """
        refmasslist = Path(refmasslist) if isinstance(refmasslist, str) else refmasslist
        
        if not refmasslist.exists():
            raise FileExistsError("File does not exist: %s" % refmasslist)


        with refmasslist.open('r') as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(1024))
            delimiter=dialect.delimiter

        if isinstance(refmasslist, S3Path):
            # data = self.file_location.open('rb').read()
            data = BytesIO(refmasslist.open('rb').read())
        
        else:
            data = refmasslist

        df_ref = pd.read_csv(data,sep=delimiter,header=None,skiprows=1)
        
        df_ref = df_ref.rename({0:'Formula',
                            1:'m/z',
                            2:'Charge',
                            3:'Form2'},axis=1)
        
        df_ref.sort_values(by='m/z',ascending=False)
        print("Reference mass list loaded - "+str(len(df_ref))+ " calibration masses loaded.")
        
        return df_ref

    def gen_ref_mass_list_from_assigned(self, mass_spectrum,min_conf=0.7):
        
        """
        This function will generate a ref mass dataframe object 
        from an assigned corems mass spec obj
        using assigned masses above a certain minimum confidence threshold

        Parameters
        ----------
        mass_spectrum : corems mass spec obj
            processed assigned mass spectrum.
        min_conf : float, optional
            minimum confidence score. The default is 0.7.

        Returns
        -------
        df_ref : pandas dataframe
            reference mass list - based on calculated masses.

        """
        df = mass_spectrum.to_dataframe()
        df = df[df['Isotopologue Similarity']>min_conf]
        df_ref = pd.DataFrame(columns=['m/z'])
        df_ref['m/z'] = df['Calculated m/z']
        print("Reference mass list generated - "+str(len(df_ref))+ " calibration masses.")
        return df_ref

    def find_calibration_points(self, mass_spectrum, df_ref,
                                calib_ppm_error_threshold=(-1,1),
                                calib_snr_threshold=5):
        """
        function to find calibration points in the mass spectrum based on the reference
        mass list. 
        
        Parameters
        ----------
        mass_spectrum : corems mass spec obj
            mass spectrum to be calibrated.
        df_ref : pandas dataframe
            reference mass list for recalibration.
        calib_ppm_error_threshold : float, optional
            ppm error for finding calibration masses in the spectrum. The default is 1.
        calib_snr_threshold : float, optional
            snr threshold for finding calibration masses in the spectrum. The default is 5.

        Returns
        -------
        imzmeas : list of ints
            indexes of measured peaks to use in mass calibration.
        mzrefs : list of floats
            reference mz values of found calibration points.

        """

        min_calib_ppm_error = calib_ppm_error_threshold[0]
        max_calib_ppm_error = calib_ppm_error_threshold[1]

        df_raw = mass_spectrum.to_dataframe()
        
        imzmeas = [] 
        mzrefs = [] 
        
        for mzref in df_ref['m/z']:
            
            # find all peaks within a defined ppm error threshold
            tmpdf = df_raw[((df_raw['m/z']-mzref)/df_raw['m/z'])*1e6<max_calib_ppm_error]
            tmpdf = tmpdf[((tmpdf['m/z']-mzref)/tmpdf['m/z'])*1e6>min_calib_ppm_error]
            
            #tmpdf = df_raw[(abs(df_raw['m/z']-mzref)/df_raw['m/z'])*1e6 < calib_ppm_error_threshold]
            
            # optionally further subset that based on minimum S/N, RP, Peak Height
            # to ensure only valid points are utilized
            # in this example, only a S/N threshold is implemented.
            tmpdf = tmpdf[tmpdf['S/N']>calib_snr_threshold]
            
            # only use the calibration point if only one peak is within the thresholds
            # This may require some optimization of the threshold tolerances
            if len(tmpdf)==1: 
                imzmeas.append(int(tmpdf.index.values))
                mzrefs.append(mzref)
                
        # it is crucial the mass lists are in same order
        # corems likes to do masses from high to low.
        mzrefs.sort(reverse=True) 
        print(str(len(imzmeas))+" calibration points matched within thresholds.")
        return imzmeas, mzrefs

    def robust_calib(self, param, imzmeas, mzrefs, mass_spectrum, order=1):
        """
        computes the rms of m/z errors to minimize when calibrating
        This is adapted from from spike

        Parameters
        ----------
        param : list of floats
            generated by minimize function from scipy optimize.
        imzmeas : list of int
            indexes of measured peaks to use in mass calibration.
        mzrefs : list of floats
            reference mz values of found calibration points.
        mass_spectrum : corems mass spec obj
            mass spectrum to be calibrated.
        order : int, optional
            order of the recalibration function. 1 = linear, 2 = quadratic. The default is 1.

        Returns
        -------
        rmserror : float
            root mean square mass error for calibration points.

        """
        Aterm = param[0]
        Bterm = param[1]
        try:
            Cterm = param[2]
        except IndexError:
            pass
        
        # get the mspeaks from the mass spectrum object which were calibration points
        mspeaks = [mass_spectrum.mspeaks[x] for x in imzmeas]
        # get their calibrated mass values
        mspeakmzs = [x.mz_cal for x in mspeaks]
        mspeakmzs = np.array(mspeakmzs)
        
        # linear
        if order == 1:
            ref_recal_points = (Aterm * mspeakmzs) + Bterm
        # quadratic
        elif order == 2:
            ref_recal_points = (Aterm * (mspeakmzs)) + \
            (Bterm * np.power((mspeakmzs), 2) + Cterm)
        
        # sort both the calibration points (measured, recalibrated)
        ref_recal_points.sort()
        # and sort the calibration points (theoretical, predefined)
        mzrefs.sort()
        
        #calculate the ppm error for each calibration point
        error = ((ref_recal_points-mzrefs)/ref_recal_points) * 1000000
        # calculate the root mean square error - this is our target to minimize
        rmserror = np.sqrt(np.mean(error**2))
        return rmserror

    def recalibrate_mass_spectrum(self, mass_spectrum, imzmeas,mzrefs,order=1):
        
        """
        function to recalibrate the mass spectrum object
        
        Parameters
        ----------
        mass_spectrum : corems mass spec obj
            mass spectrum to be calibrated.
        imzmeas : list of int
            indexes of measured peaks to use in mass calibration.
        mzrefs : list of float
            reference mz values of found calibration points.
        order : int, optional
            order of the recalibration function. 1 = linear, 2 = quadratic. The default is 1.

        Returns
        -------
        mass_spectrum : corems mass spec obj
            mass spectrum to be calibrated.

        """
        # initialise parameters for recalibration
        # these are the 'Aterm, Bterm, Cterm'
        # as spectra are already freq->mz calibrated, these terms are very small
        # may be beneficial to formally separate them from the freq->mz terms
        if order == 1:
            Po = [1,0]
        elif order == 2:
            Po = [1,0,0]
        
        if len(imzmeas) > 2:

            minimize_method = mass_spectrum.settings.calib_minimize_method
            res =  minimize(self.robust_calib, Po, args=(imzmeas, mzrefs, mass_spectrum,order), method=minimize_method)
            
            print("minimize function completed with RMS error of: {:0.3f} ppm".format(res['fun']))
            print("minimize function performed {:1d} fn evals and {:1d} iterations".format(res['nfev'],res['nit']))
            Pn = res.x
            
            mz_exp_ms = np.array(
                [mspeak.mz_exp for mspeak in mass_spectrum])  
            
            if order == 1:
                mz_domain = (Pn[0] * mz_exp_ms) + Pn[1]
                if not mass_spectrum.is_centroid:
                    mz_profile_calc = (Pn[0] * mass_spectrum.mz_exp_profile) + Pn[1]
                
            elif order == 2:
                mz_domain = (Pn[0] * (mz_exp_ms)) + \
                    (Pn[1] * np.power((mz_exp_ms), 2) + Pn[2])

                if not mass_spectrum.is_centroid:
                    mz_profile_calc = (Pn[0] * (mass_spectrum.mz_exp_profile)) + \
                            (Pn[1] * np.power((mass_spectrum.mz_exp_profile), 2) + Pn[2])
            
            mass_spectrum.mz_cal = mz_domain
            if not mass_spectrum.is_centroid:
                mass_spectrum.mz_cal_profile = mz_profile_calc

            mass_spectrum.calibration_order = order
            mass_spectrum.calibration_points = len(mzrefs)
            mass_spectrum.calibration_RMS = float(res['fun'])
        
        return mass_spectrum

    def run(self):
        
        calib_ppm_error_threshold = self.mass_spectrum.settings.calib_sn_threshold
        max_calib_ppm_error = self.mass_spectrum.settings.max_calib_ppm_error
        min_calib_ppm_error = self.mass_spectrum.settings.min_calib_ppm_error
        calib_pol_order = self.mass_spectrum.settings.calib_pol_order

        # load reference mass list
        df_ref = self.load_ref_mass_list(self.ref_mass_list_path)
        
        # find calibration points
        imzmeas, mzrefs = self.find_calibration_points(self.mass_spectrum, df_ref,
                                    calib_ppm_error_threshold=(min_calib_ppm_error, max_calib_ppm_error),
                                    calib_snr_threshold=calib_ppm_error_threshold)

        self.recalibrate_mass_spectrum(self.mass_spectrum, imzmeas, mzrefs, order=calib_pol_order)
  
    


__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"

from corems.mass_spectrum.input.baseClass import MassListBaseClass
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile, MassSpecCentroid
from corems.molecular_formula.factory.MolecularFormulaFactory import MolecularFormula
from corems.encapsulation.constant import Labels, Atoms
from corems.encapsulation.factory.processingSetting  import DataInputSetting

class ReadCoremsMasslist(MassListBaseClass):
    '''
    The ReadCoremsMasslist object reads processed mass list data types
    and returns the mass spectrum obj with the molecular formula obj

    **Only available for centroid mass spectrum type: it will ignore the parameter **isCentroid** 
    Please see MassListBaseClass for more details
    
    '''
        
    def get_mass_spectrum(self, scan_number=0, time_index=-1, auto_process=True, loadSettings=True):
        
        dataframe = self.get_dataframe()
       
        if not set(['H/C', 'O/C', 'Heteroatom Class', 'Ion Type', 'Is Isotopologue']).issubset(dataframe.columns):
            raise ValueError("%s it is not a valid CoreMS file" % str(self.file_location))
        
        self.check_columns(dataframe.columns)
        
        dataframe.rename(columns=self.parameters.header_translate, inplace=True)
 
        polarity = dataframe['Ion Charge'].values[0]

        output_parameters = self.get_output_parameters(polarity)

        mass_spec_obj = MassSpecCentroid(dataframe.to_dict(orient='list'), output_parameters)

        if loadSettings: self.load_settings(mass_spec_obj, output_parameters)

        self.add_molecular_formula(mass_spec_obj, dataframe)
        
        return mass_spec_obj

    def add_molecular_formula(self, mass_spec_obj, dataframe):

        # check if is coreMS file
        if 'Is Isotopologue' in dataframe:

            mz_exp_df = dataframe[Labels.mz].astype(float)
            # formula_df = dataframe.loc[:, 'C':].fillna(0)
            # \.replace({b'nan':0})
            formula_df = dataframe[dataframe.columns.intersection(Atoms.atoms_order)]
            formula_df.fillna(0, inplace=True)
            formula_df.replace(b'nan', 0, inplace=True)

            ion_type_df = dataframe["Ion Type"]
            ion_charge_df = dataframe["Ion Charge"]
            is_isotopologue_df = dataframe['Is Isotopologue']

        mass_spec_mz_exp_list = mass_spec_obj.mz_exp

        for df_index, mz_exp in enumerate(mz_exp_df):

            ms_peak_index = list(mass_spec_mz_exp_list).index(float(mz_exp))

            if 'Is Isotopologue' in dataframe:

                atoms = list(formula_df.columns.astype(str))
                counts = list(formula_df.iloc[df_index].astype(int))

                formula_list = [sub[item] for item in range(len(atoms))
                                for sub in [atoms, counts]]
            if sum(counts) > 0:

                ion_type = str(Labels.ion_type_translate.get(ion_type_df[df_index]))
                mfobj = MolecularFormula(formula_list, int(ion_charge_df[df_index]), mspeak_parent=mass_spec_obj[ms_peak_index] , ion_type=ion_type)
                mfobj.is_isotopologue = bool(is_isotopologue_df[df_index])
                mass_spec_obj[ms_peak_index].add_molecular_formula(mfobj)


class ReadMassList(MassListBaseClass):
    
    '''
    The ReadCoremsMasslist object reads unprocessed mass list data types
    and returns the mass spectrum obj 
    See MassListBaseClass for details
    
    '''

    def get_mass_spectrum(self, polarity, scan=0, auto_process=True, loadSettings=True):
        '''
         The MassListBaseClass object reads mass list data types and returns the mass spectrum obj

        Parameters
        ----------
        polarity: int 
            +1 or -1 
        '''
        #delimiter = "  " or " " or  "," or "\t" etc  
        
        if self.isCentroid:

            dataframe = self.get_dataframe()
        
            self.check_columns(dataframe.columns)
                
            self.clean_data_frame(dataframe)
            
            dataframe.rename(columns=self.parameters.header_translate, inplace=True)
            
            output_parameters = self.get_output_parameters(polarity)

            mass_spec = MassSpecCentroid(dataframe.to_dict(orient='list'), output_parameters)
            
            if loadSettings: self.load_settings(mass_spec, output_parameters)
            
            return mass_spec

        else:

            dataframe = self.get_dataframe()

            self.check_columns(dataframe.columns)
            
            output_parameters = self.get_output_parameters(polarity)

            self.clean_data_frame(dataframe)
            
            dataframe.rename(columns=self.parameters.header_translate, inplace=True)

            mass_spec = MassSpecProfile(dataframe.to_dict(orient='list'), output_parameters, auto_process=auto_process)

            if loadSettings: self.load_settings(mass_spec, output_parameters)

            return mass_spec
    
    



import json

from pandas import DataFrame
import h5py
from io import BytesIO
from s3path import S3Path

from corems.encapsulation.input.parameter_from_json import _set_dict_data_ms
from corems.mass_spectrum.input.massList import ReadCoremsMasslist
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecCentroid
from corems.encapsulation.constant import Labels
from corems.encapsulation.factory.parameters import default_parameters



class ReadCoreMSHDF_MassSpectrum(ReadCoremsMasslist):
    
    '''
    The MassSpectra object contains lots of MassSpectrum

    Parameters
    ----------
    arg : str
        The arg is used for ...
    *args
        The variable arguments are used for ...
    **kwargs
        The keyword arguments are used for ...

    Attributes
    ----------
    arg : str
        This is where we store arg,
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if isinstance(self.file_location, S3Path):
            data = BytesIO(self.file_location.open('rb').read())
        else:
            data = self.file_location
        
        self.h5pydata = h5py.File(data, 'r')

        self.scans = list(self.h5pydata.keys())

        print(self.scans)
    
    def load_raw_data(self, mass_spectrum, scan_index=0):

        scan_label = self.scans[scan_index]

        mz_profile = self.h5pydata[scan_label]['raw_ms'][0]
        
        abundance_profile = self.h5pydata[scan_label]['raw_ms'][1]

        mass_spectrum.mz_exp_profile = mz_profile
        
        mass_spectrum.abundance_profile = abundance_profile

    def get_mass_spectrum(self, scan_number=0, time_index=-1, auto_process=True, load_settings=True, load_raw = True):
        
        dataframe = self.get_dataframe(scan_number, time_index=time_index)
        
        if not set(['H/C', 'O/C', 'Heteroatom Class', 'Ion Type', 'Is Isotopologue']).issubset(dataframe.columns):
            raise ValueError("%s it is not a valid CoreMS file" % str(self.file_location))
        
        dataframe.rename(columns=self.parameters.header_translate, inplace=True)
 
        polarity = dataframe['Ion Charge'].values[0]

        output_parameters = self.get_output_parameters(polarity, scan_index=scan_number)

        mass_spec_obj = MassSpecCentroid(dataframe.to_dict(orient='list'), output_parameters)

        if load_settings: self.load_settings(mass_spec_obj, scan_index=scan_number, time_index=time_index)

        if load_raw: self.load_raw_data(mass_spec_obj, scan_index=scan_number)

        self.add_molecular_formula(mass_spec_obj, dataframe)
        
        return mass_spec_obj

    #override baseclass  
    def load_settings(self, mass_spectrum, scan_index = 0, time_index = -1):

        loaded_settings = {}
        loaded_settings['MoleculaSearch'] = self.get_scan_group_attr_data(scan_index,  time_index, 'MoleculaSearchSetting')
        loaded_settings['MassSpecPeak'] = self.get_scan_group_attr_data(scan_index,  time_index, 'MassSpecPeakSetting')
        loaded_settings['MassSpectrum'] = self.get_scan_group_attr_data(scan_index, time_index, 'MassSpectrumSetting')
        loaded_settings['Transient'] = self.get_scan_group_attr_data(scan_index, time_index, 'TransientSetting')
        
        _set_dict_data_ms(loaded_settings, mass_spectrum)

    #override baseclass  
    def get_dataframe(self, scan_index = 0, time_index = -1):

        columnsLabels = self.get_scan_group_attr_data(scan_index, time_index, 'ColumnsLabels')
        
        scan_label = self.scans[scan_index]

        index_to_pull = self.get_time_index_to_pull(scan_label, time_index)
        
        corems_table_data = self.h5pydata[scan_label][index_to_pull]

        list_dict = []
        for row in corems_table_data:
            data_dict = {}
            for data_index, data in enumerate(row):
                label = columnsLabels[data_index]    
                data_dict[label] = data
            
            list_dict.append(data_dict)

        return DataFrame(list_dict)
    
    
    def get_time_index_to_pull(self, scan_label, time_index):
        
        time_data = sorted([(i, int(i)) for i in self.h5pydata[scan_label].keys() if i != 'raw_ms'], key=lambda m: m[1])
        
        index_to_pull = time_data[time_index][0]     

        return index_to_pull

    def get_high_level_attr_data(self, attr_str, ):
        
        return self.h5pydata.attrs[attr_str]
   
    def get_scan_group_attr_data(self, scan_index, time_index, attr_group, attr_srt=None):
        
        scan_label = self.scans[scan_index]

        index_to_pull = self.get_time_index_to_pull(scan_label, time_index)
        
        if attr_srt:
            
            return json.loads(self.h5pydata[scan_label][index_to_pull].attrs[attr_group])[attr_srt]
        
        else:
             
             data = self.h5pydata[scan_label][index_to_pull].attrs.get(attr_group)
             if data:
                return json.loads(data)
             else:
                return {}   
   
    def get_raw_data_attr_data(self, scan_index, attr_group, attr_str):
        
        scan_label = self.scans[scan_index]
        
        return json.loads(self.h5pydata[scan_label]['raw_ms'].attrs[attr_group])[attr_str]

    #override baseclass  
    def get_output_parameters(self, polarity, scan_index=0):
        
        d_params = default_parameters(self.file_location)
        d_params["filename_path"] = self.file_location
        d_params["scan_number"] = int(self.scans[scan_index])
        d_params['polarity'] = self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'polarity')
        d_params['rt'] =     self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'rt')
        
        d_params['tic'] =  self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'tic')
        
        d_params['mobility_scan'] =    self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'mobility_scan')
        d_params['mobility_rt'] =     self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'mobility_rt')
        d_params['Aterm'] =  self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'Aterm')
        d_params['Bterm'] =  self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'Bterm')
        d_params['Cterm'] = self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'Cterm')
        d_params['baselise_noise'] = self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'baselise_noise')
        d_params['baselise_noise_std'] = self.get_raw_data_attr_data( scan_index, 'MassSpecAttrs', 'baselise_noise_std')
        
        d_params['analyzer'] = self.get_high_level_attr_data('analyzer')
        d_params['instrument_label'] = self.get_high_level_attr_data('instrument_label')
        d_params['sample_name'] = self.get_high_level_attr_data('sample_name')

        return d_params


    
    
    


__author__ = "Yuri E. Corilo"
__date__ = "Feb 13, 2020"

from pathlib import Path
import numpy as np
from pandas import Series

from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecCentroidLowRes
from corems.mass_spectra.calc import SignalProcessing as sp


class GC_Calculations:
    
    def calibrate_ri(self, ref_dict, cal_file_path):
        
        if not self:
            
            self.process_chromatogram()

        for gcms_peak in self:
        
            gcms_peak.calc_ri(ref_dict)

        self.ri_pairs_ref = ref_dict
        if  isinstance(cal_file_path, str):
			# if obj is a string it defaults to create a Path obj, pass the S3Path if needed
            self.cal_file_path = Path(cal_file_path)
        else:
            self.cal_file_path = cal_file_path
        
    def smooth_tic(self, tic):
            
        implemented_smooth_method = self.chromatogram_settings.implemented_smooth_method
        
        pol_order = self.chromatogram_settings.savgol_pol_order

        window_len = self.chromatogram_settings.smooth_window

        window = self.chromatogram_settings.smooth_method

        return sp.smooth_signal(tic, window_len, window, pol_order, implemented_smooth_method)
    
    def centroid_detector(self, tic, rt):
        
        noise_std = self.chromatogram_settings.std_noise_threshold

        method = self.chromatogram_settings.noise_threshold_method
        
        #peak picking
        min_height = self.chromatogram_settings.peak_height_min_percent 
        min_datapoints = self.chromatogram_settings.min_peak_datapoints   
        
        # baseline detection
        max_prominence = self.chromatogram_settings.peak_max_prominence_percent 
        max_height = self.chromatogram_settings.peak_height_max_percent 
        
        peak_indexes_generator = sp.peak_detector_generator(tic, noise_std, method, rt, max_height, min_height, max_prominence, min_datapoints)

        return peak_indexes_generator
  
    def remove_outliers(self, data):
        
        from numpy import percentile
        q25, q75 = percentile(data, 25), percentile(data, 75)
        iqr = q75 - q25
        print('Percentiles: 25th=%.3f, 75th=%.3f, IQR=%.3f' % (q25, q75, iqr))
        # calculate the outlier cutoff
        cut_off = iqr * 1.5
        lower, upper = q25 - cut_off, q75 + cut_off
        # identify outliers
        outliers = [x for x in data if x < lower or x > upper]
        print('Identified outliers: %d' % len(outliers))
        # remove outliers
        nanfilled_outliers = Series([x if lower <= x <= upper else np.nan for x in data])

        return nanfilled_outliers
        
   



from os.path import basename, dirname, normpath

from matplotlib import rcParamsDefault, rcParams
from numpy import linspace

from corems.encapsulation.factory.parameters import MSParameters
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecfromFreq
from corems.transient.calc.TransientCalc import TransientCalculations
import matplotlib.pyplot as plt
from copy import deepcopy
from corems.encapsulation.input.parameter_from_json import load_and_set_parameters_class


__author__ = "Yuri E. Corilo"
__date__ = "Jun 19, 2019"

'''
fig = plt.figure()

fig.patch.set_facecolor(None)

fig.patch.set_alpha(0)
'''

class Transient(TransientCalculations):
    '''
    The ReadMassList object contains lots of MassSpectrum objs

    Parameters
    ----------
    arg : str
        The arg is used for ...
    *args
        The variable arguments are used for ...
    **kwargs
        The keyword arguments are used for ...

    Attributes
    ----------
    arg : str
        This is where we store arg,
    '''
    
    
    def __init__(self, data, d_params):

        self._transient_data = data

        self.d_params = d_params

        self.frequency_domain = None

        self.magnitude = None

        self.__set__parameters__objects(d_params)

        self.__set__transient__time()

    def __set__parameters__objects(self, d_params):

        self._full_filename_path = d_params.get("filename_path")

        self.calibration_terms = (
            d_params.get("Aterm"),
            d_params.get("Bterm"),
            d_params.get("Cterm"),
        )

        self._exc_high_freq = d_params.get("exc_high_freq")

        self._exc_low_freq = d_params.get("exc_low_freq")

        self.bandwidth = d_params.get("bandwidth")

        self.number_data_points = d_params.get("number_data_points")

        self.polarity = int(d_params.get("polarity"))

        self.location = 220

        self._parameters = deepcopy(MSParameters.transient)

    def scale_plot_size(self, factor=1.5):

        default_dpi = rcParamsDefault["figure.dpi"]
        rcParams["figure.dpi"] = default_dpi * factor

    def __set__transient__time(self):
        self.transient_time = self.cal_transient_time()

    def set_processing_parameter(
        self, apodization_method, number_of_truncations, number_of_zero_fills
    ):

        self.parameters.apodization_method = apodization_method

        self.parameters.number_of_truncations = number_of_truncations

        self.parameters.number_of_zero_fills = number_of_zero_fills

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, instance_TransientParameters):
        self._parameters = instance_TransientParameters
    
    def set_parameter_from_json(self, parameters_path):
        self._parameters = load_and_set_parameters_class('Transient', self._parameters, parameters_path=parameters_path)
    
    def get_frequency_domain(self, plot_result=True):

        if self.parameters.number_of_truncations > 0:

            new_time_domain = self.truncation(self._transient_data)

        else:

            new_time_domain = self._transient_data

        if self.parameters.apodization_method is not None:

            new_time_domain = self.apodization(new_time_domain)

        if plot_result:

            self._plot_transient(self._transient_data)

            self._plot_transient(new_time_domain)

        time_domain_y_zero_filled = self.zero_fill(new_time_domain)

        self.transient_time = self.transient_time * (
            self.parameters.number_of_zero_fills + 1
        )

        if plot_result:

            self._plot_transient(time_domain_y_zero_filled)

        return self.perform_magniture_mode_ft(time_domain_y_zero_filled)
        # return frequency_domain, magnitude

    def get_mass_spectrum(self, auto_process=True, plot_result=True,
                         keep_profile=True, auto_noise=True, noise_bayes_est=False):

        #plt.figure(figsize=(13, 8))

        frequency_domain, magnitude = self.get_frequency_domain(plot_result=plot_result)

        if plot_result:

            self._plot_frequency_domain(frequency_domain, magnitude)

        self.d_params["filename"] = self.filename
        self.d_params["dir_location"] = self.dir_location
        
        
        return MassSpecfromFreq(
            frequency_domain, magnitude, self.d_params, 
            auto_process=auto_process, keep_profile=keep_profile,
            auto_noise=auto_noise, noise_bayes_est=noise_bayes_est)

    @property
    def filename(self):

        # return dirname(self._full_filename_path)
        return basename(normpath(self._full_filename_path))

    @property
    def dir_location(self):

        return dirname(self._full_filename_path).strip(
            basename(normpath(self._full_filename_path))
        )

    @property
    def A_therm(self):

        return self.calibration_terms[0]

    @property
    def B_therm(self):

        return self.calibration_terms[1]

    @property
    def C_therm(self):

        return self.calibration_terms[2]

    def _plot_frequency_domain(self, frequency_domain, magnitude): # pragma: no cover

        self.location += 1
        plt.subplot(self.location)
        plt.plot(frequency_domain, magnitude, color="green")
        plt.xlabel("Hz")
        plt.ylabel("Magnitude")
        # reset grid location index to 0
        self.location = 220
        # plt.show()

    def _plot_transient(self, transient_data): # pragma: no cover

        self.location += 1
        # print( self.location)
        time_axis = linspace(0, self.transient_time, num=len(transient_data))
        plt.subplot(self.location)
        plt.plot(time_axis, transient_data, color="green")
        plt.xlabel("Time (s)")
        plt.ylabel("Magnitude")
        # plt.show()

    def plot_transient(self): # pragma: no cover

        # self.location +=1
        # print( self.location)
        time_axis = linspace(0, self.transient_time, num=len(self._transient_data))
        # plt.subplot(self.location)
        plt.plot(time_axis, self._transient_data, color="green")
        plt.xlabel("Time (s)")
        plt.ylabel("Magnitude")
        # plt.show()

    def plot_zerofilled_transient(self): # pragma: no cover

        new_time_domain = self.apodization(self._transient_data)
        time_domain_y_zero_filled = self.zero_fill(new_time_domain)
        self.transient_time = self.transient_time * (
            self.parameters.number_of_zero_fills + 1
        )
        time_axis = linspace(0, self.transient_time, num=len(time_domain_y_zero_filled))
        # plt.subplot(self.location)
        plt.plot(time_axis, time_domain_y_zero_filled, color="green")
        plt.xlabel("Time (s)")
        plt.ylabel("Magnitude")
        # plt.show()

    def plot_apodized_transient(self):  # pragma: no cover

        # self.location +=1
        # print( self.location)
        new_time_domain = self.apodization(self._transient_data)
        time_axis = linspace(0, self.transient_time, num=len(new_time_domain))
        # plt.subplot(self.location)
        plt.plot(time_axis, new_time_domain, color="green")
        plt.xlabel("Time (s)")
        plt.ylabel("Magnitude")
        # plt.show()

    def plot_frequency_domain(self):  # pragma: no cover

        # self.location +=1
        # plt.subplot(self.location)
        frequency_domain, magnitude = self.get_frequency_domain(plot_result=False)
        plt.plot(frequency_domain / 1000, magnitude, color="green")
        plt.xlabel("KHz")
        plt.ylabel("Magnitude")
        # plt.show()
